//////////////////////////////////////////////////////////////////////////
//
//  Copyright (c) 2013, Image Engine Design Inc. All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//
//      * Redistributions of source code must retain the above
//        copyright notice, this list of conditions and the following
//        disclaimer.
//
//      * Redistributions in binary form must reproduce the above
//        copyright notice, this list of conditions and the following
//        disclaimer in the documentation and/or other materials provided with
//        the distribution.
//
//      * Neither the name of John Haddon nor the names of
//        any other contributors to this software may be used to endorse or
//        promote products derived from this software without specific prior
//        written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
//  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
//  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
//  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
//  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
//  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
//  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//////////////////////////////////////////////////////////////////////////

#ifndef GAFFER_METADATA_H
#define GAFFER_METADATA_H

#include "Gaffer/CatchingSignalCombiner.h"
#include "Gaffer/Export.h"

#include "IECore/Data.h"
#include "IECore/InternedString.h"
#include "IECore/StringAlgo.h"

#include "boost/signals.hpp"

#include <functional>

namespace Gaffer
{

IE_CORE_FORWARDDECLARE( GraphComponent )
IE_CORE_FORWARDDECLARE( Node )
IE_CORE_FORWARDDECLARE( Plug )

/// The Metadata class provides a registry of metadata for the different types
/// of Nodes and Plugs. This metadata assists in creating UIs and can be used to
/// generate documentation. Metadata can consist of either static values represented
/// as IECore::Data, or can be computed dynamically.
class GAFFER_API Metadata
{

	public :

		/// Type for a singal emitted when new metadata is registered.
		typedef boost::signal<void ( IECore::InternedString target, IECore::InternedString key ), CatchingSignalCombiner<void> > ValueChangedSignal;
		/// Type for a signal emitted when new node metadata is registered. The
		/// node argument will be null when generic (rather than per-instance)
		/// metadata is registered.
		typedef boost::signal<void ( IECore::TypeId nodeTypeId, IECore::InternedString key, Gaffer::Node *node ), CatchingSignalCombiner<void> > NodeValueChangedSignal;
		/// Type for a signal emitted when new plug metadata is registered. The
		/// plug argument will be null when generic (rather than per-instance)
		/// metadata is registered.
		typedef boost::signal<void ( IECore::TypeId nodeTypeId, const IECore::StringAlgo::MatchPattern &plugPath, IECore::InternedString key, Gaffer::Plug *plug ), CatchingSignalCombiner<void> > PlugValueChangedSignal;

		typedef std::function<IECore::ConstDataPtr ()> ValueFunction;
		typedef std::function<IECore::ConstDataPtr ( const Node *node )> NodeValueFunction;
		typedef std::function<IECore::ConstDataPtr ( const Plug *plug )> PlugValueFunction;

		/// Value registration
		/// ==================

		/// Registers a static value.
		static void registerValue( IECore::InternedString target, IECore::InternedString key, IECore::ConstDataPtr value );
		/// Registers a dynamic value. Each time the data is retrieved, the ValueFunction will
		/// be called to compute it.
		static void registerValue( IECore::InternedString target, IECore::InternedString key, ValueFunction value );

		/// Registers a static metadata value for the specified node type.
		static void registerValue( IECore::TypeId nodeTypeId, IECore::InternedString key, IECore::ConstDataPtr value );
		/// Registers a dynamic metadata value for the specified node type. Each time the data is retrieved, the
		/// NodeValueFunction will be called to compute it.
		static void registerValue( IECore::TypeId nodeTypeId, IECore::InternedString key, NodeValueFunction value );

		/// Registers a static metadata value for plugs with the specified path on the specified node type.
		static void registerValue( IECore::TypeId nodeTypeId, const IECore::StringAlgo::MatchPattern &plugPath, IECore::InternedString key, IECore::ConstDataPtr value );
		/// Registers a dynamic metadata value for the specified plug. Each time the data is retrieved, the
		/// PlugValueFunction will be called to compute it.
		static void registerValue( IECore::TypeId nodeTypeId, const IECore::StringAlgo::MatchPattern &plugPath, IECore::InternedString key, PlugValueFunction value );

		/// Registers a metadata value specific to a single instance - this will take precedence over any
		/// values registered above. If persistent is true, the value will be preserved across script save/load and cut/paste.
		/// \undoable
		static void registerValue( GraphComponent *target, IECore::InternedString key, IECore::ConstDataPtr value, bool persistent = true );

		/// Registration queries
		/// ====================

		/// Fills the keys vector with keys for all values registered with the methods above.
		static void registeredValues( IECore::InternedString target, std::vector<IECore::InternedString> &keys );
		/// Fills the keys vector with keys for all values registered for the specified graphComponent.
		/// If instanceOnly is true, then only the values registered for that exact instance are returned.
		/// If persistentOnly is true, then non-persistent instance values are ignored.
		static void registeredValues( const GraphComponent *target, std::vector<IECore::InternedString> &keys, bool instanceOnly = false, bool persistentOnly = false );

		/// Value retrieval
		/// ===============

		/// Retrieves a value, returning null if none exists.
		template<typename T=IECore::Data>
		static typename T::ConstPtr value( IECore::InternedString target, IECore::InternedString key );
		template<typename T=IECore::Data>
		static typename T::ConstPtr value( const GraphComponent *target, IECore::InternedString key, bool instanceOnly = false );

		/// Value deregistration
		/// ====================

		static void deregisterValue( IECore::InternedString target, IECore::InternedString key );
		static void deregisterValue( IECore::TypeId nodeTypeId, IECore::InternedString key );
		static void deregisterValue( IECore::TypeId nodeTypeId, const IECore::StringAlgo::MatchPattern &plugPath, IECore::InternedString key );

		/// \undoable
		static void deregisterValue( GraphComponent *target, IECore::InternedString key );

		/// Utilities
		/// =========

		/// Lists all node descendants of "root" with the specified metadata key.
		/// If instanceOnly is true the search is restricted to instance metadata.
		static std::vector<Node*> nodesWithMetadata( GraphComponent *root, IECore::InternedString key, bool instanceOnly = false );

		/// Lists all plug descendants of "root" with the specified metadata key.
		/// If instanceOnly is true the search is restricted to instance metadata.
		static std::vector<Plug*> plugsWithMetadata( GraphComponent *root, IECore::InternedString key, bool instanceOnly = false );

		/// Signals
		/// =======
		///
		/// These are emitted when the Metadata has been changed with one
		/// of the register*() methods. If dynamic metadata is registered
		/// with a NodeValueFunction or PlugValueFunction then it is the
		/// responsibility of the registrant to manually emit the signals
		/// when necessary.
		static ValueChangedSignal &valueChangedSignal();
		static NodeValueChangedSignal &nodeValueChangedSignal();
		static PlugValueChangedSignal &plugValueChangedSignal();

		/// Deprecated
		/// ==============

		/// \deprecated
		static void registerNodeValue( IECore::TypeId nodeTypeId, IECore::InternedString key, IECore::ConstDataPtr value );
		/// \deprecated
		static void registerNodeValue( IECore::TypeId nodeTypeId, IECore::InternedString key, NodeValueFunction value );
		/// \deprecated
		static void registerNodeValue( Node *node, IECore::InternedString key, IECore::ConstDataPtr value, bool persistent = true );
		/// \deprecated
		static void registeredNodeValues( const Node *node, std::vector<IECore::InternedString> &keys, bool inherit = true, bool instanceOnly = false, bool persistentOnly = false );
		/// \deprecated
		template<typename T>
		static typename T::ConstPtr nodeValue( const Node *node, IECore::InternedString key, bool inherit = true, bool instanceOnly = false );
		/// \deprecated
		static void deregisterNodeValue( IECore::TypeId nodeTypeId, IECore::InternedString key );
		/// \deprecated
		static void deregisterNodeValue( Node *node, IECore::InternedString key );

		/// \deprecated
		static void registerPlugValue( IECore::TypeId nodeTypeId, const IECore::StringAlgo::MatchPattern &plugPath, IECore::InternedString key, IECore::ConstDataPtr value );
		/// \deprecated
		static void registerPlugValue( IECore::TypeId nodeTypeId, const IECore::StringAlgo::MatchPattern &plugPath, IECore::InternedString key, PlugValueFunction value );
		/// \deprecated
		static void registerPlugValue( Plug *plug, IECore::InternedString key, IECore::ConstDataPtr value, bool persistent = true );
		/// \deprecated
		static void registeredPlugValues( const Plug *plug, std::vector<IECore::InternedString> &keys, bool inherit = true, bool instanceOnly = false, bool persistentOnly = false );
		/// \deprecated
		template<typename T>
		static typename T::ConstPtr plugValue( const Plug *plug, IECore::InternedString key, bool inherit = true, bool instanceOnly = false );
		/// \deprecated
		static void deregisterPlugValue( IECore::TypeId nodeTypeId, const IECore::StringAlgo::MatchPattern &plugPath, IECore::InternedString key );
		/// \deprecated
		static void deregisterPlugValue( Plug *plug, IECore::InternedString key );

		/// \deprecated
		static void registerNodeDescription( IECore::TypeId nodeTypeId, const std::string &description );
		/// \deprecated
		static void registerNodeDescription( IECore::TypeId nodeTypeId, NodeValueFunction description );
		/// \deprecated
		static std::string nodeDescription( const Node *node, bool inherit = true );

		/// \deprecated
		static void registerPlugDescription( IECore::TypeId nodeTypeId, const IECore::StringAlgo::MatchPattern &plugPath, const std::string &description );
		/// \deprecated
		static void registerPlugDescription( IECore::TypeId nodeTypeId, const IECore::StringAlgo::MatchPattern &plugPath, PlugValueFunction description );
		/// \deprecated
		static std::string plugDescription( const Plug *plug, bool inherit = true );

	private :

		/// Per-instance Metadata is stored as a mapping from GraphComponent * to the
		/// metadata values, and needs to be removed when the instance dies. Currently
		/// there is no callback when a RefCounted object passes away, so we must rely
		/// on the destructors for Node and Plug to call clearInstanceMetadata() for us.
		/// \todo This situation isn't particularly satisfactory - if we introduced
		/// weak pointers and destruction callbacks for RefCounted objects then we could
		/// tidy this up.
		friend class Node;
		friend class Plug;
		static void clearInstanceMetadata( const GraphComponent *graphComponent );

		static IECore::ConstDataPtr valueInternal( IECore::InternedString target, IECore::InternedString key );
		static IECore::ConstDataPtr valueInternal( const GraphComponent *target, IECore::InternedString key, bool instanceOnly );
		static IECore::ConstDataPtr nodeValueInternal( const Node *node, IECore::InternedString key, bool inherit, bool instanceOnly );
		static IECore::ConstDataPtr plugValueInternal( const Plug *plug, IECore::InternedString key, bool inherit, bool instanceOnly );

};

} // namespace Gaffer

#include "Gaffer/Metadata.inl"

#endif // GAFFER_METADATA_H
